/*-
 * Copyright (C) @BABOLO  2008 Mar 16 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2008 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ipclass_btail.c,v 1.2 2010/07/30 14:08:13 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "ipclassify.h"

size_t
newnode(ipclass_btail *opts) {
    size_t node;

    node = opts->ptree++;
    if  (!opts->ip_tree) {
        opts->ltree = (opts->filin / 3 + 256) & ~(size_t)0xFF;
        opts->ip_tree = malloc(opts->ltree * sizeof(ipclassify_er));
    } else if (opts->ptree >= opts->ltree) {
        ifBLIN_QV2(opts->flags) warnx("realloc %u+", opts->ltree);
        opts->ltree += 256;
        opts->ip_tree = reallocf(opts->ip_tree, opts->ltree * sizeof(ipclassify_er));
    }
    if  (!opts->ip_tree) {
        ifBLIN_QV1(opts->flags) warnx("No RAM for %u ipclassify_ers", opts->ltree);
        node = ~(u_int32_t)0;
    }
    return(node);
}

size_t
#ifdef IPTRACE
ipclass_bbit(ipclass_btail *opts, int bit, u_int32_t ip) {
#else
ipclass_bbit(ipclass_btail *opts, int bit) {
#endif
    int c;
    u_int32_t node, be, bn;

    ifBLIN_QV3(opts->flags)
#ifdef IPTRACE
        fprintf( stderr, "+ %08X bit=%u %08X [%u]=%c\n"
               , opts->flags, bit, ip, opts->tpoint, opts->plain[opts->tpoint]
               );
#else
        fprintf( stderr, "+ %08X bit=%u [%u]=%c\n"
               , opts->flags, bit, opts->tpoint, opts->plain[opts->tpoint]
               );
#endif
    if  (opts->tpoint >= opts->filin) {
        ifBLIN_QV1(opts->flags) warnx("End of plain tree");
        node = ~(u_int32_t)0;
        goto out;
    }
    if  (bit > 32) {
        ifBLIN_QV1(opts->flags) warnx("Plain tree depth too high");
        node = ~(u_int32_t)0;
        goto out;
    }
    c = opts->plain[opts->tpoint++];
    if  (c == 'f') {
        ifBLIN_QV5(opts->flags) fprintf(stderr, "f %u\n", bit);
        node = opts->false;
        opts->flags |= IPCLASS_F;
    } else if (c == 't') {
        ifBLIN_QV5(opts->flags) fprintf(stderr, "t %u\n", bit);
        if  (!(opts->flags & IPCLASS_T)) {
            ifBLIN_QV5(opts->flags) fprintf(stderr, "new t %u\n", opts->true);
            opts->true = newnode(opts);
            opts->ip_tree[opts->true].mask = 0;
            opts->ip_tree[opts->true].value = 0;
            opts->ip_tree[opts->true].eq = opts->eq;
            opts->ip_tree[opts->true].neq = opts->neq;
            opts->flags |= IPCLASS_T;
        }
        node = opts->true;
    } else if (c == 'q') {
        ifBLIN_QV5(opts->flags) fprintf(stderr, "q %u\n", bit);
        node = newnode(opts);
        opts->ip_tree[node].mask = 0;
        opts->ip_tree[node].value = 0;
        opts->ip_tree[node].eq = 0;
        opts->ip_tree[node].neq = 0;
#ifdef IPTRACE
        be = ipclass_bbit(opts, bit + 1, ip);
        bn = ipclass_bbit(opts, bit + 1, ip | (0x80000000 >> bit));
#else
        be = ipclass_bbit(opts, bit + 1);
        bn = ipclass_bbit(opts, bit + 1);
#endif

        opts->ip_tree[node].mask = 0x80000000 >> bit;
        opts->ip_tree[node].value = 0;
        opts->ip_tree[node].eq = be;
        opts->ip_tree[node].neq = bn;
    } else {
        ifBLIN_QV1(opts->flags) warnx("Ill sym in plain tree at %u", opts->tpoint - 1);
        node = ~(u_int32_t)0;
        goto out;
    }
out:
    ifBLIN_QV3(opts->flags) {
        fprintf(stderr, "- %u %08X [%u]", node, opts->flags, opts->tpoint);
        if (node != ~(size_t)0) {
            fprintf( stderr, ",[%u]={%08X,%08X,%u,%u}\n"
                   , node + opts->stree
                   , opts->ip_tree[node].mask
                   , opts->ip_tree[node].value
                   , opts->ip_tree[node].eq
                   , opts->ip_tree[node].neq
                   );
        } else {
            fprintf(stderr, "\n");
    }   }
    return(node);
}

static void
usage(int ex) {
    fprintf( stderr
           , "ipclassify @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "ipclassify [... ] iwclass owclass out[ in]\n"
             " -c - constant tail\n"
             " -v[v[v..]] - verbose\n"
             " -q - quiet\n"
           );
    exit(ex);
}

int
main(int argc, char **argv) {
    int ex = EX_OK, c;
    mife_descriptor *fin = NULL;
    ipclass_btail *opts;
    struct stat sb;

    if  (!(opts = malloc(sizeof(ipclass_btail)))) {
        ifBLIN_QV1(opts->flags) warn("No RAM #1");
        ex = EX_UNAVAILABLE;
        goto out;
    }
    bzero(opts, sizeof(ipclass_btail));
    opts->fout = -1;
    opts->flags = BLIN_VER1;
    opts->false = ~(u_int32_t)0;
    while ((c = getopt(argc, argv, "chqv?")) > 0) {
        switch (c) {
        case 'c':
            opts->flags |= IPCLASS_C;
            break;
        case 'q':
            opts->flags &= ~BLIN_VERMASK;
            break;
        case 'v':
            BLIN_VERBOSE(opts->flags);
            break;
        case 'h': usage(EX_OK);
        default: usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc > 0 && argv && argv[0] && argv[0][0]) {
        opts->eq = strtol(argv[0], NULL, 0);
        opts->flags |= IPCLASS_I;
    }
    if  (argc > 1 && argv && argv[1] && argv[1][0]) {
        opts->neq = strtol(argv[1], NULL, 0);
        opts->flags |= IPCLASS_O;
    }
    if  (~opts->flags & (IPCLASS_I | IPCLASS_O)) {
        ifBLIN_QV1(opts->flags) warnx("Must have in and out classes");
        ex = EX_NOINPUT;
        goto out;
    }
    if  (  argc > 2 && argv && argv[2] && argv[2][0]
        && !(opts->fout = open(argv[2], O_WRONLY | O_APPEND | O_CREAT | O_EXLOCK))
        ) {
        ifBLIN_QV1(opts->flags) warn("Cant\'t create %s", argv[2]);
        ex = EX_CANTCREAT;
        goto out;
    }
    if  (fstat(opts->fout, &sb) < 0) {
        ifBLIN_QV1(opts->flags) warn("Cant\'t stat %s", argv[2]);
        ex = EX_IOERR;
        goto out;
    }
    if  (S_ISREG(sb.st_mode)) {
        opts->stree = sb.st_size / sizeof(ipclassify_er);
            if  (sb.st_size != opts->stree * sizeof(ipclassify_er)) {
            ifBLIN_QV1(opts->flags) warnx("Broken file for update to %s", argv[2]);
            ex = EX_DATAERR;
            goto out;
    }   }
    if  (opts->flags & IPCLASS_C) {
        opts->ip_tree = malloc(sizeof(ipclassify_er));
        opts->ltree = 1;
        opts->ip_tree[0].mask = 0;
        opts->ip_tree[0].value = 0;
        opts->ip_tree[0].eq = opts->eq;
        opts->ip_tree[0].neq = opts->neq;
        opts->ptree = 1;
        opts->flags |= IPCLASS_P;
    } else {
        if  (argc > 3 && argv && argv[3] && argv[3][0] && !(fin = mife_open(0, argv[3]))) {
            ifBLIN_QV1(opts->flags) warn("Can't open: %s", argv[3]);
            ex = EX_NOINPUT;
            goto out;
        }
        if  ((opts->filin = mife_read(fin, 0, 0)) < 0) {
            ifBLIN_QV1(opts->flags) warn("Can't read: %s", argv[1]);
            ex = EX_IOERR;
            goto out;
        }
        if  (!(opts->plain = mife_get(fin, 0))) {
            ifBLIN_QV1(opts->flags) warn("Can't get: %s", argv[1]);
            ex = EX_IOERR;
            goto out;
        }
#ifdef IPTRACE
        switch (ipclass_bbit(opts, 0, 0)) {
#else
        switch (ipclass_bbit(opts, 0)) {
#endif
        case 0:
            if  (opts->ltree) {
                for (c = 0; c < opts->ptree; c++) {
                    if  (opts->ip_tree[c].mask) {
                        if  (opts->ip_tree[c].eq == opts->false) {
                            opts->ip_tree[c].eq = opts->ptree + opts->stree;
                        } else {
                            opts->ip_tree[c].eq += opts->stree;
                        }
                        if  (opts->ip_tree[c].neq == opts->false) {
                            opts->ip_tree[c].neq = opts->ptree + opts->stree;
                        } else {
                            opts->ip_tree[c].neq += opts->stree;
            }   }   }   }
            opts->flags |= IPCLASS_P;
            break;
        case ~(u_int32_t)0:
            break;
        default:
            ifBLIN_QV1(opts->flags) warnx("Broken tree");
            ex = EX_DATAERR;
            goto out;
    }   }
    if  (  (opts->flags & IPCLASS_P)
        && mife_writ(opts->fout, opts->ip_tree, opts->ptree * sizeof(ipclassify_er)) < 0
        ) {
        ifBLIN_QV1(opts->flags) warn("Can't write output");
        ex = EX_IOERR;
        goto out;
    }
out:
    exit(ex);
}
