/*-
 * Copyright (C) @BABOLO  2010 Jul 30
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: traf-testpol.c,v 1.1 2010/07/30 14:08:13 babolo Exp $ 
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include "ipclassify.h"

ipclassify_er *ip_tree;

void print_ip(FILE *f, u_int32_t ip)
{   fprintf(f, "%3u", (ip >> 24) & 0xFF);
    fprintf(f, ".%u", (ip >> 16) & 0xFF);
    fprintf(f, ".%u", (ip >> 8) & 0xFF);
    fprintf(f, ".%u", ip & 0xFF);
}

int
readipaddr( char **iptext
          , u_int32_t *ipaddr
          )
/* reads in as IP address and converts to u_int_32 */
{   long long j;
    u_int32_t i;
    long k;
    char *inn = NULL;

    j = 0LL; *ipaddr = 0;
    for (i = 0; i < 4;)
    {   if  (i)
        {   if  (**iptext == '.') (*iptext)++;
            else if (strspn(*iptext, " \t\r\n") == 0)                                       return(33);
            else break;
        }else if (strspn(*iptext, "0123456789") == 0)                                       return(34);
        inn = *iptext;
        j = strtoll(*iptext, iptext, 0);
        if  (  j == LLONG_MAX || j == LLONG_MIN)                                            return(-1);
        if  (  (j & 0xffffffff00000000LL) != ((j < 0LL) ? 0xffffffff00000000LL : 0LL)
            || *iptext == inn
            )                                                                               return(35);
        j &= 0x00000000ffffffffLL;
        if  (j < 0x0100LL)
        {   *ipaddr <<= 8;
            *ipaddr |= (u_int32_t)j;
            i += 1;
        }else if (j < 0x010000LL)
        {   *ipaddr <<= 16;
            *ipaddr |= (u_int32_t)j;
            i += 2;
        }else if (j < 0x01000000LL)
        {   *ipaddr <<= 24;
            *ipaddr |= (u_int32_t)j;
            i += 3;
        }else
        {   *ipaddr = (u_int32_t)j;
            i += 4;
        }
#ifdef DEBUG
        fprintf( stderr, "%1d, %08X, %08X, %s~\n", i, *ipaddr, (u_int32_t)j, *iptext);
#endif
    }
    if  (i > 4)                                                                             return(36);
    i = 4 - i;
    switch (i)
    {case 1:
        *ipaddr <<= 8;
        if   (j < 0x0100LL)        *ipaddr &= 0xffff0000;
        else if (j < 0x010000LL)   *ipaddr &= 0xff000000;
        else if (j < 0x01000000LL) *ipaddr = 0;
        *ipaddr |= j;
        break;
     case 2:
        *ipaddr <<= 16;
        if   (j < 0x0100LL)      *ipaddr &= 0xff000000;
        else if (j < 0x010000LL) *ipaddr = 0;
        *ipaddr |= j;
        break;
    }
    k = 32L;
    if  (**iptext == '/')
    {   (*iptext)++;
        inn = *iptext;
        k = strtol(*iptext, iptext, 0);
        if  (  k == LONG_MAX || k == LONG_MIN)                                              return(-2);
        if  (k < 0L || k > 32L || *iptext == inn)                                           return(38);
    }else if (**iptext) return(37);
    return(k);
/* return codes
 * -2 - strtol (mask)
 * -1 - strtoll (*ipaddr)
 * 0 .. 32 result mask length
 * 33 - Illegal syntax
 * 34 - Illegal syntax
 * 35 - Out of range
 * 36 - Out of range
 * 37 -    
 * 38 - Mask out of range
 */
}

void usage(int status)
{   fprintf( stderr
           , "traf-testpol @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "traf-testpol PolicyFile IPAddr\n"
           );
    exit(status);
}

int main(int argc, char **argv)
{   int ifile, p;
    struct stat sb;
    size_t len_tree;
    u_int32_t ipaddr;

    if (argc != 3) usage(EX_USAGE);
    if  ((ifile = open(argv[1], O_RDONLY)) < 0)              err(EX_NOINPUT, "Can't open %s", argv[1]);
    if  (fstat(ifile, &sb) < 0)                               err(EX_NOINPUT, "Can't stat input file");
    ip_tree = (ipclassify_er*)mmap( NULL
                                  , sb.st_size
                                  , PROT_READ
                                  , 0
                                  , ifile
                                  , 0
                                  );
    if  ((void*)ip_tree == MAP_FAILED)                        err(EX_NOINPUT, "Can't mmap input file");
    len_tree = sb.st_size / sizeof(ipclassify_er);
    close(ifile);
    p = readipaddr(&argv[2], &ipaddr);
    if  (p < 0 || p > 32)                                             errx(EX_USAGE, "IP input error");
    if  (p != 32)                                                errx(EX_USAGE, "Use /32 mask please");

    for (p = 0; ip_tree[p].mask != 0;)
    {   if  ((ip_tree[p].mask & ipaddr) == ip_tree[p].value)
        {   printf( "%5d == (m%08X v%08X i%08X)\n"
                  , p, ip_tree[p].mask & ipaddr, ip_tree[p].value, ipaddr
                  );
            p = ip_tree[p].eq;
        }else
        {   printf( "%5d <> (m%08X v%08X i%08X)\n"
                  , p, ip_tree[p].mask & ipaddr, ip_tree[p].value, ipaddr
                  );
            p = ip_tree[p].neq;
    }   }

    printf("%5d %5d %5d\n", p, ip_tree[p].eq, ip_tree[p].neq);
    exit(EX_OK);
}
